@@ -15,6 +15,7 @@ from twisted.internet.protocol import BaseProtocol
 
 from gnutls.connection import ClientSession, ServerSession, ServerSessionFactory
 from gnutls.connection import X509Credentials as _X509Credentials
+from gnutls.constants import SHUT_RDWR, SHUT_WR
 from gnutls.errors import *
 
 
@@ -62,11 +63,10 @@ class X509Credentials(_X509Credentials):
     verify_period = None
 
     def verify_callback(self, peer_cert, preverify_status=None):
-        """Verifies peer certificate and raises an exception if it cannot be accepted"""
+        """Verifies the peer certificate and raises an exception if it cannot be accepted"""
         if isinstance(preverify_status, Exception):
             raise preverify_status
-        self.check_certificate(peer_cert)
-        
+        self.check_certificate(peer_cert, cert_name='peer certificate')
 
 
 class TLSMixin:
@@ -93,16 +93,47 @@ class TLSMixin:
         except GNUTLSError, e:
             return e
 
-    def _postLoseConnection(self):
+    def _sendCloseReason(self, reason):
         try:
-            self.socket.bye()
-        except OperationWouldBlock, e:
-            # Since we do not continue to use the connection after closing TLS
-            # we do not need to wait for the bye notification acknowledgement.
+            self.socket.send_alert(reason)
+        except OperationInterrupted:
+            self._sendCloseReason(reason)
+
+    def _sendCloseAlert(self, how=SHUT_RDWR):
+        try:
+            self.socket.bye(how)
+        except OperationInterrupted:
+            self._sendCloseAlert(how)
+
+    def closeTLSSession(self, reason):
+        try:
+            self._sendCloseReason(reason)
+            self._sendCloseAlert(SHUT_RDWR)
+        except Exception, e:
             pass
-        except GNUTLSError, e:
+
+    def _postLoseConnection(self):
+        self.closeTLSSession(self._close_reason)
+        return self._close_reason
+
+    def endTLSWrite(self):
+        self.stopWriting()
+        try:
+            self._sendCloseAlert(SHUT_WR)
+        except OperationWouldBlock, e:
+            if self.socket.interrupted_while_writing:
+                self.startWriting()
+                return
+        except Exception, e:
             return e
-        return getattr(self, '_close_reason', main.CONNECTION_DONE)
+        del self.doWrite
+
+    def _closeWriteConnection(self):
+        self.doWrite = self.endTLSWrite
+        result = self.endTLSWrite()
+        if isinstance(result, Exception):
+            return result
+        return tcp.Connection._closeWriteConnection(self)
 
 
 class TLSClient(TLSMixin, tcp.Client):
@@ -154,12 +185,17 @@ class TLSClient(TLSMixin, tcp.Client):
                 self.startWriting()
             return
         except GNUTLSError, e:
-            self.failIfNotConnected(err = error.getConnectError(str(e)))
+            self.failIfNotConnected(err = e)
             return
         
         try:
             self._verifyPeer()
+        except GNUTLSError, e:
+            self.closeTLSSession(e)
+            self.failIfNotConnected(err = e)
+            return
         except Exception, e:
+            self.closeTLSSession(e)
             self.failIfNotConnected(err = error.getConnectError(str(e)))
             return
         
@@ -293,6 +329,7 @@ class TLSPort(tcp.Port):
 
     def _preMakeConnection(self, transport):
         transport.protocol.makeConnection = lambda *args: None
+        transport.protocol.transport = transport ## because we may call connectionLost without connectionMade
         transport.startTLS()
         return tcp.Port._preMakeConnection(self, transport)
 
